<!DOCTYPE html>
<html>
<head>
    <meta charset="UTF-8">
    <link href="sedrila.css" rel="stylesheet">
    <link href="local.css" rel="stylesheet">
    <link href="codehilite.css" rel="stylesheet">
    <link rel="icon" type="image/png" sizes="32x32" href="favicon-32x32.png">
    <title>{{title}}</title>
</head>
<body>
  <script src="sidebar.js" defer></script>
  <div id="preheader"></div>
{% block header %}
  <header role="banner" class="header">
    <div class="header_in">
      <button type="button" class="toggle" id="toggle">
        <span class="toggle"></span>
      </button>
      <a href="index.html">{{sitetitle}}</a>
    </div>
  </header>
{% endblock %}
  <nav class="sidebar" id="sidebar">
      {{ toc }}
  </nav>
  <div class="main" id="main">
   {% block breadcrumb %}
      {% if breadcrumb %}
        {{breadcrumb}}
      {% endif %}
   {% endblock %}
   <main>
    {% block titleline %}
      {% if title %}
        <h1>{{title}}</h1>
      {% endif %}
    {% endblock %}
    {% block linkslist_top %}
       <!--linkslist_top appears only in tasks-->
    {% endblock %}
    {% block mainpart %}
      <div>
          {{ content }}
      </div>
    {% endblock %} 
    <!--- subsequent tasks: -->
    {% block linkslist_bottom %}
       <!--linkslist_bottom appears only in tasks-->
    {% endblock %}
   </main>  
  {% block footer %}
   <footer>
      <p class="copyright">Generated by <a href="https://sedrila.readthedocs.io/en/latest/">sedrila</a></p>
   </footer>
  {% endblock %}
  </div>
</body>
</html>
