{% extends "base.html" %}
{% block linkslist_top %}
    {{linkslist_top}}
{% endblock %}
{% block mainpart %}
  <div id="taskbody" class="pagetype-task pagetype-task-difficulty{{ part.difficulty}}">
      {{ content }}
  </div>
{% endblock %}
{% block linkslist_bottom %}
    {{linkslist_bottom}}
{% endblock %}
